# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for msg editor interactions.."""

from .... import msgeditor
from ....tests import TestCaseWithTransport
from ... import commitfromnews


class TestRegisters(TestCaseWithTransport):
    def test_registered_at_import(self):
        self.assertTrue(commitfromnews._registered)

    def test_register_registers_for_commit_message_template(self):
        commitfromnews._registered = False
        # Registers only within the plugin
        commitfromnews.register()
        self.assertLength(1, msgeditor.hooks["commit_message_template"])
        self.assertTrue(commitfromnews._registered)
